@@ -1,10 +1,10 @@
 
 /* pngrutil.c - utilities to read a png file
 
-   libpng 1.0 beta 1 - version 0.71
+   libpng 1.0 beta 3 - version 0.89
    For conditions of distribution and use, see copyright notice in png.h
-   Copyright (c) 1995 Guy Eric Schalnat, Group 42, Inc.
-   June 26, 1995
+   Copyright (c) 1995, 1996 Guy Eric Schalnat, Group 42, Inc.
+   May 25, 1996
    */
 
 #define PNG_INTERNAL
@@ -12,7 +12,7 @@
 
 /* grab an uint 32 from a buffer */
 png_uint_32
-png_get_uint_32(png_byte *buf)
+png_get_uint_32(png_bytep buf)
 {
    png_uint_32 i;
 
@@ -26,19 +26,19 @@ png_get_uint_32(png_byte *buf)
 
 /* grab an uint 16 from a buffer */
 png_uint_16
-png_get_uint_16(png_byte *buf)
+png_get_uint_16(png_bytep buf)
 {
    png_uint_16 i;
 
-   i = ((png_uint_16)(*buf) << 8) +
-      (png_uint_16)(*(buf + 1));
+   i = (png_uint_16)(((png_uint_16)(*buf) << 8) +
+      (png_uint_16)(*(buf + 1)));
 
    return i;
 }
 
 /* read data, and run it through the crc */
 void
-png_crc_read(png_struct *png_ptr, png_byte *buf, png_uint_32 length)
+png_crc_read(png_structp png_ptr, png_bytep buf, png_uint_32 length)
 {
    png_read_data(png_ptr, buf, length);
    png_calculate_crc(png_ptr, buf, length);
@@ -46,7 +46,7 @@ png_crc_read(png_struct *png_ptr, png_byte *buf, png_uint_32 length)
 
 /* skip data, but calcuate the crc anyway */
 void
-png_crc_skip(png_struct *png_ptr, png_uint_32 length)
+png_crc_skip(png_structp png_ptr, png_uint_32 length)
 {
    png_uint_32 i;
 
@@ -64,13 +64,16 @@ png_crc_skip(png_struct *png_ptr, png_uint_32 length)
 
 /* read and check the IDHR chunk */
 void
-png_handle_IHDR(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_IHDR(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_byte buf[13];
    png_uint_32 width, height;
    int bit_depth, color_type, compression_type, filter_type;
    int interlace_type;
 
+   if (png_ptr->mode != PNG_BEFORE_IHDR)
+      png_error(png_ptr, "Out of place IHDR");
+
    /* check the length */
    if (length != 13)
       png_error(png_ptr, "Invalid IHDR chunk");
@@ -87,43 +90,43 @@ png_handle_IHDR(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
    /* check for width and height valid values */
    if (width == 0 || height == 0)
-      png_error(png_ptr, "Invalid Width or Height Found");
+      png_error(png_ptr, "Invalid image size in IHDR");
 
    /* check other values */
    if (bit_depth != 1 && bit_depth != 2 &&
       bit_depth != 4 && bit_depth != 8 &&
       bit_depth != 16)
-      png_error(png_ptr, "Invalid Bit Depth Found");
+      png_error(png_ptr, "Invalid bit depth in IHDR");
 
    if (color_type < 0 || color_type == 1 ||
       color_type == 5 || color_type > 6)
-      png_error(png_ptr, "Invalid Color Type Found");
+      png_error(png_ptr, "Invalid color type in IHDR");
 
    if (color_type == PNG_COLOR_TYPE_PALETTE &&
       bit_depth == 16)
-      png_error(png_ptr, "Found Invalid Color Type and Bit Depth Combination");
+      png_error(png_ptr, "Invalid color type and bit depth combination in IHDR");
 
    if ((color_type == PNG_COLOR_TYPE_RGB ||
       color_type == PNG_COLOR_TYPE_GRAY_ALPHA ||
       color_type == PNG_COLOR_TYPE_RGB_ALPHA) &&
       bit_depth < 8)
-      png_error(png_ptr, "Found Invalid Color Type and Bit Depth Combination");
+      png_error(png_ptr, "Invalid color type and bit depth in IHDR");
 
    if (interlace_type > 1)
-      png_error(png_ptr, "Found Invalid Interlace Value");
+      png_error(png_ptr, "Invalid interlace method in IHDR");
 
    if (compression_type > 0)
-      png_error(png_ptr, "Found Invalid Compression Value");
+      png_error(png_ptr, "Invalid compression method in IHDR");
 
    if (filter_type > 0)
-      png_error(png_ptr, "Found Invalid Filter Value");
+      png_error(png_ptr, "Invalid filter method in IHDR");
 
    /* set internal variables */
    png_ptr->width = width;
    png_ptr->height = height;
-   png_ptr->bit_depth = bit_depth;
-   png_ptr->interlaced = interlace_type;
-   png_ptr->color_type = color_type;
+   png_ptr->bit_depth = (png_byte)bit_depth;
+   png_ptr->interlaced = (png_byte)interlace_type;
+   png_ptr->color_type = (png_byte)color_type;
 
    /* find number of channels */
    switch (png_ptr->color_type)
@@ -143,27 +146,54 @@ png_handle_IHDR(png_struct *png_ptr, png_info *info, png_uint_32 length)
          break;
    }
    /* set up other useful info */
-   png_ptr->pixel_depth = png_ptr->bit_depth *
-      png_ptr->channels;
+   png_ptr->pixel_depth = (png_byte)(png_ptr->bit_depth *
+      png_ptr->channels);
    png_ptr->rowbytes = ((png_ptr->width *
       (png_uint_32)png_ptr->pixel_depth + 7) >> 3);
    /* call the IHDR callback (which should just set up info) */
    png_read_IHDR(png_ptr, info, width, height, bit_depth,
       color_type, compression_type, filter_type, interlace_type);
+
+   png_ptr->mode |= PNG_HAVE_IHDR;
 }
 
 /* read and check the palette */
 void
-png_handle_PLTE(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_PLTE(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    int num, i;
-   png_color *palette;
+   png_colorp palette;
+
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before PLTE");
+   else if (png_ptr->mode & PNG_HAVE_PLTE)
+      png_error(png_ptr, "Multiple PLTE");
+
+#if !defined(PNG_READ_OPT_PLTE_SUPPORTED)
+   if (png_ptr->color_type != PNG_COLOR_TYPE_PALETTE)
+   {
+      png_crc_skip(png_ptr, length);
+      return;
+   }
+#endif
 
    if (length % 3)
-      png_error(png_ptr, "Invalid Palette Chunk");
+   {
+      if (png_ptr->color_type != PNG_COLOR_TYPE_PALETTE)
+      {
+         png_warning(png_ptr, "Invalid palette chunk");
+         png_crc_skip(png_ptr, length);
+         return;
+      }
+      else
+      {
+         png_error(png_ptr, "Invalid palette chunk");
+      }
+   }
 
    num = (int)length / 3;
-   palette = (png_color *)png_malloc(png_ptr, num * sizeof (png_color));
+   palette = (png_colorp)png_large_malloc(png_ptr, num * sizeof (png_color));
+   png_ptr->do_free |= PNG_FREE_PALETTE;
    for (i = 0; i < num; i++)
    {
       png_byte buf[3];
@@ -175,19 +205,29 @@ png_handle_PLTE(png_struct *png_ptr, png_info *info, png_uint_32 length)
       palette[i].blue = buf[2];
    }
    png_ptr->palette = palette;
-   png_ptr->num_palette = num;
+   png_ptr->num_palette = (png_uint_16)num;
    png_read_PLTE(png_ptr, info, palette, num);
+
+   png_ptr->mode |= PNG_HAVE_PLTE;
 }
 
+#if defined(PNG_READ_gAMA_SUPPORTED)
 void
-png_handle_gAMA(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_gAMA(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_uint_32 igamma;
    float gamma;
    png_byte buf[4];
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before gAMA");
+   else if (png_ptr->mode & PNG_HAVE_PLTE)
+      /* Should be an error, but we can cope with it */
+      png_warning(png_ptr, "Out of place gAMA chunk");
+
    if (length != 4)
    {
+      png_warning(png_ptr, "Incorrect gAMA chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -202,13 +242,23 @@ png_handle_gAMA(png_struct *png_ptr, png_info *info, png_uint_32 length)
    png_read_gAMA(png_ptr, info, gamma);
    png_ptr->gamma = gamma;
 }
+#endif
 
+#if defined(PNG_READ_sBIT_SUPPORTED)
 void
-png_handle_sBIT(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_sBIT(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    int slen;
    png_byte buf[4];
 
+   buf[0] = buf[1] = buf[2] = buf[3] = 0;
+
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before sBIT");
+   else if (png_ptr->mode & PNG_HAVE_PLTE)
+      /* Should be an error, but we can cope with it */
+      png_warning(png_ptr, "Out of place sBIT chunk");
+
    if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
       slen = 3;
    else
@@ -216,6 +266,7 @@ png_handle_sBIT(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
    if (length != (png_uint_32)slen)
    {
+      png_warning(png_ptr, "Incorrect sBIT chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -235,16 +286,25 @@ png_handle_sBIT(png_struct *png_ptr, png_info *info, png_uint_32 length)
    }
    png_read_sBIT(png_ptr, info, &(png_ptr->sig_bit));
 }
+#endif
 
+#if defined(PNG_READ_cHRM_SUPPORTED)
 void
-png_handle_cHRM(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_cHRM(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_byte buf[4];
    png_uint_32 v;
    float white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y;
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before sBIT");
+   else if (png_ptr->mode & PNG_HAVE_PLTE)
+      /* Should be an error, but we can cope with it */
+      png_warning(png_ptr, "Missing PLTE before cHRM");
+
    if (length != 32)
    {
+      png_warning(png_ptr, "Incorrect cHRM chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -284,21 +344,33 @@ png_handle_cHRM(png_struct *png_ptr, png_info *info, png_uint_32 length)
    png_read_cHRM(png_ptr, info,
       white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y);
 }
+#endif
 
+#if defined(PNG_READ_tRNS_SUPPORTED)
 void
-png_handle_tRNS(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_tRNS(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before tRNS");
+
    if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
    {
-      if (length > png_ptr->num_palette)
+      if (!(png_ptr->mode & PNG_HAVE_PLTE))
       {
+         /* Should be an error, but we can cope with it */
+         png_warning(png_ptr, "Missing PLTE before tRNS");
+      }
+      else if (length > png_ptr->num_palette)
+      {
+         png_warning(png_ptr, "Incorrect tRNS chunk length");
          png_crc_skip(png_ptr, length);
          return;
       }
 
-      png_ptr->trans = png_malloc(png_ptr, length);
+      png_ptr->trans = (png_bytep)png_large_malloc(png_ptr, length);
+      png_ptr->do_free |= PNG_FREE_TRANS;
       png_crc_read(png_ptr, png_ptr->trans, length);
-      png_ptr->num_trans = (int)length;
+      png_ptr->num_trans = (png_uint_16)length;
    }
    else if (png_ptr->color_type == PNG_COLOR_TYPE_RGB)
    {
@@ -306,6 +378,7 @@ png_handle_tRNS(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
       if (length != 6)
       {
+         png_warning(png_ptr, "Incorrect tRNS chunk length");
          png_crc_skip(png_ptr, length);
          return;
       }
@@ -322,6 +395,7 @@ png_handle_tRNS(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
       if (length != 2)
       {
+         png_warning(png_ptr, "Incorrect tRNS chunk length");
          png_crc_skip(png_ptr, length);
          return;
       }
@@ -331,18 +405,34 @@ png_handle_tRNS(png_struct *png_ptr, png_info *info, png_uint_32 length)
       png_ptr->trans_values.gray = png_get_uint_16(buf);
    }
    else
-      png_error(png_ptr, "Invalid tRNS chunk");
+   {
+      png_warning(png_ptr, "tRNS chunk not allowed with alpha channel");
+      png_crc_skip(png_ptr, length);
+      return;
+   }
 
    png_read_tRNS(png_ptr, info, png_ptr->trans, png_ptr->num_trans,
       &(png_ptr->trans_values));
 }
+#endif
 
+#if defined(PNG_READ_bKGD_SUPPORTED)
 void
-png_handle_bKGD(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_bKGD(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    int truelen;
    png_byte buf[6];
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before bKGD");
+   else if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE &&
+            !(png_ptr->mode & PNG_HAVE_PLTE))
+   {
+      png_warning(png_ptr, "Missing PLTE before bKGD");
+      png_crc_skip(png_ptr, length);
+      return;
+   }
+
    if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
       truelen = 1;
    else if (png_ptr->color_type & PNG_COLOR_MASK_COLOR)
@@ -352,15 +442,30 @@ png_handle_bKGD(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
    if (length != (png_uint_32)truelen)
    {
+      png_warning(png_ptr, "Incorrect bKGD chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
 
    png_crc_read(png_ptr, buf, length);
+   /* We convert the index value into RGB components so that we can allow
+    * arbitrary RGB values for background when we have transparency, and
+    * so it is easy to determine the RGB values of the background color
+    * from the info_ptr. */
    if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
+   {
       png_ptr->background.index = buf[0];
+      png_ptr->background.red = (png_uint_16)png_ptr->palette[buf[0]].red;
+      png_ptr->background.green = (png_uint_16)png_ptr->palette[buf[0]].green;
+      png_ptr->background.blue = (png_uint_16)png_ptr->palette[buf[0]].blue;
+   }
    else if (!(png_ptr->color_type & PNG_COLOR_MASK_COLOR))
+   {
+      png_ptr->background.red =
+      png_ptr->background.green =
+      png_ptr->background.blue =
       png_ptr->background.gray = png_get_uint_16(buf);
+   }
    else
    {
       png_ptr->background.red = png_get_uint_16(buf);
@@ -370,20 +475,34 @@ png_handle_bKGD(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
    png_read_bKGD(png_ptr, info, &(png_ptr->background));
 }
+#endif
 
+#if defined(PNG_READ_hIST_SUPPORTED)
 void
-png_handle_hIST(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_hIST(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    int num, i;
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before hIST");
+   else if (!(png_ptr->mode & PNG_HAVE_PLTE))
+   {
+      png_warning(png_ptr, "Missing PLTE before hIST");
+      png_crc_skip(png_ptr, length);
+      return;
+   }
+
    if (length != 2 * png_ptr->num_palette)
    {
+      png_warning(png_ptr, "Incorrect hIST chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
 
    num = (int)length / 2;
-   png_ptr->hist = png_malloc(png_ptr, num * sizeof (png_uint_16));
+   png_ptr->hist = (png_uint_16p)png_large_malloc(png_ptr,
+      num * sizeof (png_uint_16));
+   png_ptr->do_free |= PNG_FREE_HIST;
    for (i = 0; i < num; i++)
    {
       png_byte buf[2];
@@ -393,16 +512,22 @@ png_handle_hIST(png_struct *png_ptr, png_info *info, png_uint_32 length)
    }
    png_read_hIST(png_ptr, info, png_ptr->hist);
 }
+#endif
 
+#if defined(PNG_READ_pHYs_SUPPORTED)
 void
-png_handle_pHYs(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_pHYs(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_byte buf[9];
    png_uint_32 res_x, res_y;
    int unit_type;
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before pHYS");
+
    if (length != 9)
    {
+      png_warning(png_ptr, "Incorrect pHYs chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -414,16 +539,22 @@ png_handle_pHYs(png_struct *png_ptr, png_info *info, png_uint_32 length)
    unit_type = buf[8];
    png_read_pHYs(png_ptr, info, res_x, res_y, unit_type);
 }
+#endif
 
+#if defined(PNG_READ_oFFs_SUPPORTED)
 void
-png_handle_oFFs(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_oFFs(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_byte buf[9];
    png_uint_32 offset_x, offset_y;
    int unit_type;
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before oFFs");
+
    if (length != 9)
    {
+      png_warning(png_ptr, "Incorrect oFFs chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -435,15 +566,21 @@ png_handle_oFFs(png_struct *png_ptr, png_info *info, png_uint_32 length)
    unit_type = buf[8];
    png_read_oFFs(png_ptr, info, offset_x, offset_y, unit_type);
 }
+#endif
 
+#if defined(PNG_READ_tIME_SUPPORTED)
 void
-png_handle_tIME(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_tIME(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
    png_byte buf[7];
    png_time mod_time;
 
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before tIME");
+
    if (length != 7)
    {
+      png_warning(png_ptr, "Incorrect tIME chunk length");
       png_crc_skip(png_ptr, length);
       return;
    }
@@ -459,40 +596,49 @@ png_handle_tIME(png_struct *png_ptr, png_info *info, png_uint_32 length)
 
    png_read_tIME(png_ptr, info, &mod_time);
 }
+#endif
 
+#if defined(PNG_READ_tEXt_SUPPORTED)
 /* note: this does not correctly handle chunks that are > 64K */
 void
-png_handle_tEXt(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_tEXt(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
-   char *key, *text;
+   png_charp key;
+   png_charp text;
 
-   text = NULL;
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before tEXt");
 
-   key = (char *)png_large_malloc(png_ptr, length + 1);
-   png_crc_read(png_ptr, (png_byte *)key, length);
+   key = (png_charp )png_large_malloc(png_ptr, length + 1);
+   png_crc_read(png_ptr, (png_bytep )key, length);
    key[(png_size_t)length] = '\0';
 
    for (text = key; *text; text++)
-      /* empty loop */ ;
+      /* empty loop to check key length */ ;
 
    if (text != key + (png_size_t)length)
       text++;
 
    png_read_tEXt(png_ptr, info, key, text, length - (text - key));
 }
+#endif
 
-/* note: this does not correctly handle chunks that are > 64K compressed */
+#if defined(PNG_READ_zTXt_SUPPORTED)
+/* note: this does not correctly handle chunks that are > 64K compressed
+   on those systems that can't malloc more than 64KB at a time. */
 void
-png_handle_zTXt(png_struct *png_ptr, png_info *info, png_uint_32 length)
+png_handle_zTXt(png_structp png_ptr, png_infop info, png_uint_32 length)
 {
-   char *key, *text;
-   int ret;
+   static char msg[] = "Error decoding zTXt chunk";
+   png_charp key;
+   png_charp text;
    png_uint_32 text_size, key_size;
 
-   text = NULL;
+   if (!(png_ptr->mode & PNG_HAVE_IHDR))
+      png_error(png_ptr, "Missing IHDR before zTXt");
 
    key = png_large_malloc(png_ptr, length + 1);
-   png_crc_read(png_ptr, (png_byte *)key, length);
+   png_crc_read(png_ptr, (png_bytep )key, length);
    key[(png_size_t)length] = '\0';
 
    for (text = key; *text; text++)
@@ -501,100 +647,112 @@ png_handle_zTXt(png_struct *png_ptr, png_info *info, png_uint_32 length)
    /* zTXt can't have zero text */
    if (text == key + (png_size_t)length)
    {
-      png_large_free(png_ptr, key);
-      return;
+      png_warning(png_ptr, "Zero length zTXt chunk");
+      text_size = 0;
    }
-
-   text++;
-
-   if (*text) /* check compression byte */
+   else if (*(++text)) /* check compression type byte */
    {
-      png_large_free(png_ptr, key);
-      return;
+      png_warning(png_ptr, "Unknown zTXt compression type");
+
+      /* Copy what we can of the error message into the text chunk */
+      text_size = length - (text - key) - 1;
+      text_size = sizeof(msg) > text_size ? text_size : sizeof(msg);
+      png_memcpy(text, msg, (png_size_t)(text_size + 1));
    }
-
-   text++;
-
-   png_ptr->zstream->next_in = (png_byte *)text;
-   png_ptr->zstream->avail_in = (uInt)(length - (text - key));
-   png_ptr->zstream->next_out = png_ptr->zbuf;
-   png_ptr->zstream->avail_out = (png_size_t)png_ptr->zbuf_size;
-
-   key_size = text - key;
-   text_size = 0;
-   text = NULL;
-
-   while (png_ptr->zstream->avail_in)
+   else
    {
-      ret = inflate(png_ptr->zstream, Z_PARTIAL_FLUSH);
-      if (ret != Z_OK && ret != Z_STREAM_END)
-      {
-         inflateReset(png_ptr->zstream);
-         png_ptr->zstream->avail_in = 0;
-         png_large_free(png_ptr, key);
-         png_large_free(png_ptr, text);
-         return;
-      }
-      if (!png_ptr->zstream->avail_out || ret == Z_STREAM_END)
-      {
-         if (!text)
-         {
-            text = png_malloc(png_ptr,
-               png_ptr->zbuf_size - png_ptr->zstream->avail_out +
-                  key_size + 1);
-            memcpy(text + (png_size_t)key_size, png_ptr->zbuf,
-               (png_size_t)(png_ptr->zbuf_size - png_ptr->zstream->avail_out));
-            memcpy(text, key, (png_size_t)key_size);
-            text_size = key_size + (png_size_t)png_ptr->zbuf_size -
-               png_ptr->zstream->avail_out;
-            *(text + (png_size_t)text_size) = '\0';
-         }
-         else
-         {
-            char *tmp;
+      text++;
 
-            tmp = text;
-            text = png_large_malloc(png_ptr, text_size +
-               png_ptr->zbuf_size - png_ptr->zstream->avail_out + 1);
-            memcpy(text, tmp, (png_size_t)text_size);
-            png_large_free(png_ptr, tmp);
-            memcpy(text + (png_size_t)text_size, png_ptr->zbuf,
-               (png_size_t)(png_ptr->zbuf_size - png_ptr->zstream->avail_out));
-            text_size += png_ptr->zbuf_size - png_ptr->zstream->avail_out;
-            *(text + (png_size_t)text_size) = '\0';
-         }
-         if (ret != Z_STREAM_END)
+      png_ptr->zstream->next_in = (png_bytep )text;
+      png_ptr->zstream->avail_in = (uInt)(length - (text - key));
+      png_ptr->zstream->next_out = png_ptr->zbuf;
+      png_ptr->zstream->avail_out = (png_size_t)png_ptr->zbuf_size;
+
+      key_size = text - key;
+      text_size = 0;
+      text = NULL;
+
+      while (png_ptr->zstream->avail_in)
+      {
+         int ret;
+
+         ret = inflate(png_ptr->zstream, Z_PARTIAL_FLUSH);
+         if (ret != Z_OK && ret != Z_STREAM_END)
          {
-            png_ptr->zstream->next_out = png_ptr->zbuf;
-            png_ptr->zstream->avail_out = (uInt)png_ptr->zbuf_size;
+            if (png_ptr->zstream->msg)
+               png_warning(png_ptr, png_ptr->zstream->msg);
+            else
+               png_warning(png_ptr, "zTXt decompression error");
+            inflateReset(png_ptr->zstream);
+            png_ptr->zstream->avail_in = 0;
+
+            if (!text)
+            {
+               text_size = key_size + sizeof(msg) + 1;
+               text = (png_charp)png_large_malloc(png_ptr, text_size);
+               png_memcpy(text, key, (png_size_t)key_size);
+            }
+
+            text[text_size - 1] = '\0';
+
+            /* Copy what we can of the error message into the text chunk */
+            text_size = length - (text - key) - 1;
+            text_size = sizeof(msg) > text_size ? text_size : sizeof(msg);
+            png_memcpy(text + key_size, msg, (png_size_t)(text_size + 1));
+            break;
+         }
+         if (!png_ptr->zstream->avail_out || ret == Z_STREAM_END)
+         {
+            if (!text)
+            {
+               text = (png_charp)png_large_malloc(png_ptr,
+                  png_ptr->zbuf_size - png_ptr->zstream->avail_out +
+                     key_size + 1);
+               png_memcpy(text + (png_size_t)key_size, png_ptr->zbuf,
+                  (png_size_t)(png_ptr->zbuf_size - png_ptr->zstream->avail_out));
+               png_memcpy(text, key, (png_size_t)key_size);
+               text_size = key_size + (png_size_t)png_ptr->zbuf_size -
+                  png_ptr->zstream->avail_out;
+               *(text + (png_size_t)text_size) = '\0';
+            }
+            else
+            {
+               png_charp tmp;
+
+               tmp = text;
+               text = png_large_malloc(png_ptr, text_size +
+                  png_ptr->zbuf_size - png_ptr->zstream->avail_out + 1);
+               png_memcpy(text, tmp, (png_size_t)text_size);
+               png_large_free(png_ptr, tmp);
+               png_memcpy(text + (png_size_t)text_size, png_ptr->zbuf,
+                  (png_size_t)(png_ptr->zbuf_size - png_ptr->zstream->avail_out));
+               text_size += png_ptr->zbuf_size - png_ptr->zstream->avail_out;
+               *(text + (png_size_t)text_size) = '\0';
+            }
+            if (ret != Z_STREAM_END)
+            {
+               png_ptr->zstream->next_out = png_ptr->zbuf;
+               png_ptr->zstream->avail_out = (uInt)png_ptr->zbuf_size;
+            }
+            else
+            {
+               break;
+            }
          }
       }
-      else
-      {
-         break;
-      }
 
-      if (ret == Z_STREAM_END)
-         break;
-   }
+      inflateReset(png_ptr->zstream);
+      png_ptr->zstream->avail_in = 0;
 
-   inflateReset(png_ptr->zstream);
-   png_ptr->zstream->avail_in = 0;
-
-   if (ret != Z_STREAM_END)
-   {
       png_large_free(png_ptr, key);
-      png_large_free(png_ptr, text);
-      return;
+      key = text;
+      text += (png_size_t)key_size;
+      text_size -= key_size;
    }
 
-   png_large_free(png_ptr, key);
-   key = text;
-   text += (png_size_t)key_size;
-   text_size -= key_size;
-
    png_read_zTXt(png_ptr, info, key, text, text_size, 0);
 }
+#endif
 
 /* Combines the row recently read in with the previous row.
    This routine takes care of alpha and transparency if requested.
@@ -608,12 +766,12 @@ png_handle_zTXt(png_struct *png_ptr, png_info *info, png_uint_32 length)
    you want all pixels to be combined, pass 0xff (255) in mask.
 */
 void
-png_combine_row(png_struct *png_ptr, png_byte *row,
+png_combine_row(png_structp png_ptr, png_bytep row,
    int mask)
 {
    if (mask == 0xff)
    {
-      memcpy(row, png_ptr->row_buf + 1,
+      png_memcpy(row, png_ptr->row_buf + 1,
          (png_size_t)((png_ptr->width *
          png_ptr->row_info.pixel_depth + 7) >> 3));
    }
@@ -623,8 +781,8 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
       {
          case 1:
          {
-            png_byte *sp;
-            png_byte *dp;
+            png_bytep sp;
+            png_bytep dp;
             int m;
             int shift;
             png_uint_32 i;
@@ -640,7 +798,7 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
                {
                   value = (*sp >> shift) & 0x1;
                   *dp &= (png_byte)((0x7f7f >> (7 - shift)) & 0xff);
-                  *dp |= (value << shift);
+                  *dp |= (png_byte)(value << shift);
                }
 
                if (shift == 0)
@@ -661,8 +819,8 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
          }
          case 2:
          {
-            png_byte *sp;
-            png_byte *dp;
+            png_bytep sp;
+            png_bytep dp;
             int m;
             int shift;
             png_uint_32 i;
@@ -678,7 +836,7 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
                {
                   value = (*sp >> shift) & 0x3;
                   *dp &= (png_byte)((0x3f3f >> (6 - shift)) & 0xff);
-                  *dp |= (value << shift);
+                  *dp |= (png_byte)(value << shift);
                }
 
                if (shift == 0)
@@ -698,8 +856,8 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
          }
          case 4:
          {
-            png_byte *sp;
-            png_byte *dp;
+            png_bytep sp;
+            png_bytep dp;
             int m;
             int shift;
             png_uint_32 i;
@@ -715,7 +873,7 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
                {
                   value = (*sp >> shift) & 0xf;
                   *dp &= (png_byte)((0xf0f >> (4 - shift)) & 0xff);
-                  *dp |= (value << shift);
+                  *dp |= (png_byte)(value << shift);
                }
 
                if (shift == 0)
@@ -735,8 +893,8 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
          }
          default:
          {
-            png_byte *sp;
-            png_byte *dp;
+            png_bytep sp;
+            png_bytep dp;
             png_uint_32 i;
             int pixel_bytes, m;
 
@@ -749,7 +907,7 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
             {
                if (m & mask)
                {
-                  memcpy(dp, sp, pixel_bytes);
+                  png_memcpy(dp, sp, pixel_bytes);
                }
 
                sp += pixel_bytes;
@@ -766,8 +924,9 @@ png_combine_row(png_struct *png_ptr, png_byte *row,
    }
 }
 
+#if defined(PNG_READ_INTERLACING_SUPPORTED)
 void
-png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
+png_do_read_interlace(png_row_infop row_info, png_bytep row, int pass)
 {
    if (row && row_info)
    {
@@ -779,7 +938,7 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
       {
          case 1:
          {
-            png_byte *sp, *dp;
+            png_bytep sp, dp;
             int sshift, dshift;
             png_byte v;
             png_uint_32 i;
@@ -791,7 +950,7 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
             dshift = 7 - (int)((final_width + 7) & 7);
             for (i = row_info->width; i; i--)
             {
-               v = (*sp >> sshift) & 0x1;
+               v = (png_byte)((*sp >> sshift) & 0x1);
                for (j = 0; j < png_pass_inc[pass]; j++)
                {
                   *dp &= (png_byte)((0x7f7f >> (7 - dshift)) & 0xff);
@@ -816,7 +975,7 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
          }
          case 2:
          {
-            png_byte *sp, *dp;
+            png_bytep sp, dp;
             int sshift, dshift;
             png_byte v;
             png_uint_32 i, j;
@@ -827,11 +986,11 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
             dshift = (png_size_t)((3 - ((final_width + 3) & 3)) << 1);
             for (i = row_info->width; i; i--)
             {
-               v = (*sp >> sshift) & 0x3;
+               v = (png_byte)((*sp >> sshift) & 0x3);
                for (j = 0; j < png_pass_inc[pass]; j++)
                {
                   *dp &= (png_byte)((0x3f3f >> (6 - dshift)) & 0xff);
-                  *dp |= (v << dshift);
+                  *dp |= (png_byte)(v << dshift);
                   if (dshift == 6)
                   {
                      dshift = 0;
@@ -852,7 +1011,7 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
          }
          case 4:
          {
-            png_byte *sp, *dp;
+            png_bytep sp, dp;
             int sshift, dshift;
             png_byte v;
             png_uint_32 i;
@@ -864,11 +1023,11 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
             dshift = (png_size_t)((1 - ((final_width + 1) & 1)) << 2);
             for (i = row_info->width; i; i--)
             {
-               v = (*sp >> sshift) & 0xf;
+               v = (png_byte)((*sp >> sshift) & 0xf);
                for (j = 0; j < png_pass_inc[pass]; j++)
                {
                   *dp &= (png_byte)((0xf0f >> (4 - dshift)) & 0xff);
-                  *dp |= (v << dshift);
+                  *dp |= (png_byte)(v << dshift);
                   if (dshift == 4)
                   {
                      dshift = 0;
@@ -889,7 +1048,7 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
          }
          default:
          {
-            png_byte *sp, *dp;
+            png_bytep sp, dp;
             png_byte v[8];
             png_uint_32 i;
             int j;
@@ -901,10 +1060,10 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
             dp = row + (png_size_t)((final_width - 1) * pixel_bytes);
             for (i = row_info->width; i; i--)
             {
-               memcpy(v, sp, pixel_bytes);
+               png_memcpy(v, sp, pixel_bytes);
                for (j = 0; j < png_pass_inc[pass]; j++)
                {
-                  memcpy(dp, v, pixel_bytes);
+                  png_memcpy(dp, v, pixel_bytes);
                   dp -= pixel_bytes;
                }
                sp -= pixel_bytes;
@@ -917,10 +1076,11 @@ png_do_read_interlace(png_row_info *row_info, png_byte *row, int pass)
          (png_uint_32)row_info->pixel_depth + 7) >> 3);
    }
 }
+#endif
 
 void
-png_read_filter_row(png_row_info *row_info, png_byte *row,
-   png_byte *prev_row, int filter)
+png_read_filter_row(png_structp png_ptr, png_row_infop row_info, png_bytep row,
+   png_bytep prev_row, int filter)
 {
    switch (filter)
    {
@@ -930,8 +1090,8 @@ png_read_filter_row(png_row_info *row_info, png_byte *row,
       {
          png_uint_32 i;
          int bpp;
-         png_byte *rp;
-         png_byte *lp;
+         png_bytep rp;
+         png_bytep lp;
 
          bpp = (row_info->pixel_depth + 7) / 8;
          for (i = (png_uint_32)bpp, rp = row + bpp, lp = row;
@@ -944,8 +1104,8 @@ png_read_filter_row(png_row_info *row_info, png_byte *row,
       case 2:
       {
          png_uint_32 i;
-         png_byte *rp;
-         png_byte *pp;
+         png_bytep rp;
+         png_bytep pp;
 
          for (i = 0, rp = row, pp = prev_row;
             i < row_info->rowbytes; i++, rp++, pp++)
@@ -958,9 +1118,9 @@ png_read_filter_row(png_row_info *row_info, png_byte *row,
       {
          png_uint_32 i;
          int bpp;
-         png_byte *rp;
-         png_byte *pp;
-         png_byte *lp;
+         png_bytep rp;
+         png_bytep pp;
+         png_bytep lp;
 
          bpp = (row_info->pixel_depth + 7) / 8;
          for (i = 0, rp = row, pp = prev_row;
@@ -980,10 +1140,10 @@ png_read_filter_row(png_row_info *row_info, png_byte *row,
       {
          int bpp;
          png_uint_32 i;
-         png_byte *rp;
-         png_byte *pp;
-         png_byte *lp;
-         png_byte *cp;
+         png_bytep rp;
+         png_bytep pp;
+         png_bytep lp;
+         png_bytep cp;
 
          bpp = (row_info->pixel_depth + 7) / 8;
          for (i = 0, rp = row, pp = prev_row,
@@ -1019,12 +1179,13 @@ png_read_filter_row(png_row_info *row_info, png_byte *row,
          break;
       }
       default:
+         png_error(png_ptr, "Bad adaptive filter type");
          break;
    }
 }
 
 void
-png_read_finish_row(png_struct *png_ptr)
+png_read_finish_row(png_structp png_ptr)
 {
    png_ptr->row_number++;
    if (png_ptr->row_number < png_ptr->num_rows)
@@ -1033,7 +1194,7 @@ png_read_finish_row(png_struct *png_ptr)
    if (png_ptr->interlaced)
    {
       png_ptr->row_number = 0;
-      memset(png_ptr->prev_row, 0, (png_size_t)png_ptr->rowbytes + 1);
+      png_memset(png_ptr->prev_row, 0, (png_size_t)png_ptr->rowbytes + 1);
       do
       {
          png_ptr->pass++;
@@ -1054,13 +1215,15 @@ png_read_finish_row(png_struct *png_ptr)
             if (!(png_ptr->num_rows))
                continue;
          }
+         if (png_ptr->transformations & PNG_INTERLACE)
+            break;
       } while (png_ptr->iwidth == 0);
 
       if (png_ptr->pass < 7)
          return;
    }
 
-   if (!png_ptr->zlib_finished)
+   if (!(png_ptr->flags & PNG_FLAG_ZLIB_FINISHED))
    {
       char extra;
       int ret;
@@ -1080,14 +1243,14 @@ png_read_finish_row(png_struct *png_ptr)
                crc = png_get_uint_32(buf);
                if (((crc ^ 0xffffffffL) & 0xffffffffL) !=
                   (png_ptr->crc & 0xffffffffL))
-                  png_error(png_ptr, "Bad CRC value");
+                  png_warning(png_ptr, "Bad CRC value");
 
                png_read_data(png_ptr, buf, 4);
                png_ptr->idat_size = png_get_uint_32(buf);
                png_reset_crc(png_ptr);
 
                png_crc_read(png_ptr, buf, 4);
-               if (memcmp(buf, png_IDAT, 4))
+               if (png_memcmp(buf, png_IDAT, 4))
                   png_error(png_ptr, "Not enough image data");
 
             }
@@ -1104,11 +1267,12 @@ png_read_finish_row(png_struct *png_ptr)
             if (!(png_ptr->zstream->avail_out) || png_ptr->zstream->avail_in ||
                png_ptr->idat_size)
                png_error(png_ptr, "Extra compressed data");
-            png_ptr->mode = PNG_AT_LAST_IDAT;
+            png_ptr->mode |= PNG_AT_LAST_IDAT;
             break;
          }
          if (ret != Z_OK)
-            png_error(png_ptr, "Compression Error");
+            png_error(png_ptr, png_ptr->zstream->msg ? png_ptr->zstream->msg :
+                      "Decompression Error");
 
          if (!(png_ptr->zstream->avail_out))
             png_error(png_ptr, "Extra compressed data");
@@ -1122,11 +1286,11 @@ png_read_finish_row(png_struct *png_ptr)
 
    inflateReset(png_ptr->zstream);
 
-   png_ptr->mode = PNG_AT_LAST_IDAT;
+   png_ptr->mode |= PNG_AT_LAST_IDAT;
 }
 
 void
-png_read_start_row(png_struct *png_ptr)
+png_read_start_row(png_structp png_ptr)
 {
    int max_pixel_depth;
    png_uint_32 rowbytes;
@@ -1157,12 +1321,13 @@ png_read_start_row(png_struct *png_ptr)
 
    max_pixel_depth = png_ptr->pixel_depth;
 
+#if defined(PNG_READ_PACK_SUPPORTED)
    if ((png_ptr->transformations & PNG_PACK) && png_ptr->bit_depth < 8)
-   {
       max_pixel_depth = 8;
-   }
+#endif
 
-   if (png_ptr->transformations & (PNG_EXPAND | PNG_PACK))
+#if defined(PNG_READ_EXPAND_SUPPORTED)
+   if (png_ptr->transformations & PNG_EXPAND)
    {
       if (png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
       {
@@ -1187,13 +1352,17 @@ png_read_start_row(png_struct *png_ptr)
          }
       }
    }
+#endif
 
-   if (png_ptr->transformations & PNG_RGBA)
+#if defined(PNG_READ_FILLER_SUPPORTED)
+   if (png_ptr->transformations & (PNG_FILLER))
    {
       if (max_pixel_depth < 32)
          max_pixel_depth = 32;
    }
+#endif
 
+#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
    if (png_ptr->transformations & PNG_GRAY_TO_RGB)
    {
       if ((png_ptr->num_trans && (png_ptr->transformations & PNG_EXPAND)) ||
@@ -1212,6 +1381,7 @@ png_read_start_row(png_struct *png_ptr)
             max_pixel_depth = 48;
       }
    }
+#endif
 
    /* align the width on the next larger 8 pixels.  Mainly used
       for interlacing */
@@ -1222,22 +1392,18 @@ png_read_start_row(png_struct *png_ptr)
       1 + ((max_pixel_depth + 7) >> 3);
 #ifdef PNG_MAX_MALLOC_64K
    if (rowbytes > 65536L)
-      png_error(png_ptr, "This image requires a row greater then 64KB");
+      png_error(png_ptr, "This image requires a row greater than 64KB");
 #endif
-   png_ptr->row_buf = (png_byte *)png_large_malloc(png_ptr, rowbytes);
+   png_ptr->row_buf = (png_bytep )png_large_malloc(png_ptr, rowbytes);
 
 #ifdef PNG_MAX_MALLOC_64K
    if (png_ptr->rowbytes + 1 > 65536L)
-      png_error(png_ptr, "This image requires a row greater then 64KB");
+      png_error(png_ptr, "This image requires a row greater than 64KB");
 #endif
-   png_ptr->prev_row = png_large_malloc(png_ptr,
-      png_ptr->rowbytes + 1);
+   png_ptr->prev_row = png_large_malloc(png_ptr, png_ptr->rowbytes + 1);
 
-   memset(png_ptr->prev_row, 0, (png_size_t)png_ptr->rowbytes + 1);
+   png_memset(png_ptr->prev_row, 0, (png_size_t)png_ptr->rowbytes + 1);
 
-   png_ptr->row_init = 1;
-
-   /* if we have to do any modifications of values for the transformations,
-      do them here */
+   png_ptr->flags |= PNG_FLAG_ROW_INIT;
 }
 
